"""Dict and List utilities"""

import itertools
from collections import UserDict
from typing import Any, Callable, Dict, List, Optional, Union

from .frictionless_wrapper import get_frictionless_resource
from .typehint import ListDict, ListTuple, ListTupleDict


def merge_two_dictionaries(
    a: Dict, b: Dict, merge_lists: bool = False
) -> Dict:
    """Merges b into a and returns merged result

    NOTE: tuples and arbitrary objects are not handled as it is totally ambiguous what should happen

    Args:
        a (Dict): dictionary to merge into
        b (Dict): dictionary to merge from
        merge_lists (bool): Whether to merge lists (True) or replace lists (False). Default is False.

    Returns:
        Dict: Merged dictionary
    """
    key = None
    # ## debug output
    # sys.stderr.write(f"DEBUG: {b} to {a}\n")
    try:
        if a is None or isinstance(a, (str, int, float)):
            # border case for first run or if a is a primitive
            a = b
        elif isinstance(a, list):
            # lists can be appended or replaced
            if isinstance(b, list):
                if merge_lists:
                    # merge lists
                    a.extend(b)
                else:
                    # replace list
                    a = b
            else:
                # append to list
                a.append(b)
        elif isinstance(a, (dict, UserDict)):
            # dicts must be merged
            if isinstance(b, (dict, UserDict)):
                for key in b:
                    if key in a:
                        a[key] = merge_two_dictionaries(
                            a[key], b[key], merge_lists=merge_lists
                        )
                    else:
                        a[key] = b[key]
            else:
                raise ValueError(
                    f'Cannot merge non-dict "{b}" into dict "{a}"'
                )
        else:
            raise ValueError(f'NOT IMPLEMENTED "{b}" into "{a}"')
    except TypeError as e:
        raise ValueError(
            f'TypeError "{e}" in key "{key}" when merging "{b}" into "{a}"'
        )
    return a


def merge_dictionaries(
    dicts: ListTuple[Dict], merge_lists: bool = False
) -> Dict:
    """Merges all dictionaries in dicts into a single dictionary and returns result

    Args:
        dicts (ListTuple[Dict]): Dictionaries to merge into the first one in the list
        merge_lists (bool): Whether to merge lists (True) or replace lists (False). Default is False.

    Returns:
        Dict: Merged dictionary

    """
    dict1 = dicts[0]
    for other_dict in dicts[1:]:
        merge_two_dictionaries(dict1, other_dict, merge_lists=merge_lists)
    return dict1


def dict_diff(d1: Dict, d2: Dict, no_key: str = "<KEYNOTFOUND>") -> Dict:
    """Compares two dictionaries

    Args:
        d1 (Dict): First dictionary to compare
        d2 (Dict): Second dictionary to compare
        no_key (str): What value to use if key is not found Defaults to '<KEYNOTFOUND>'.

    Returns:
        Dict: Comparison dictionary

    """
    d1keys = set(d1.keys())
    d2keys = set(d2.keys())
    both = d1keys & d2keys
    diff = {k: (d1[k], d2[k]) for k in both if d1[k] != d2[k]}
    diff.update({k: (d1[k], no_key) for k in d1keys - both})
    diff.update({k: (no_key, d2[k]) for k in d2keys - both})
    return diff


def dict_of_lists_add(dictionary: Dict, key: Any, value: Any) -> None:
    """Add value to a list in a dictionary by key

    Args:
        dictionary (Dict): Dictionary to which to add values
        key (Any): Key within dictionary
        value (Any): Value to add to list in dictionary

    Returns:
        None

    """
    list_objs = dictionary.get(key, list())
    list_objs.append(value)
    dictionary[key] = list_objs


def dict_of_sets_add(dictionary: Dict, key: Any, value: Any) -> None:
    """Add value to a set in a dictionary by key

    Args:
        dictionary (Dict): Dictionary to which to add values
        key (Any): Key within dictionary
        value (Any): Value to add to set in dictionary

    Returns:
        None

    """
    set_objs = dictionary.get(key, set())
    set_objs.add(value)
    dictionary[key] = set_objs


def dict_of_dicts_add(
    dictionary: Dict, parent_key: Any, key: Any, value: Any
) -> None:
    """Add key value pair to a dictionary within a dictionary by key

    Args:
        dictionary (Dict): Dictionary to which to add values
        parent_key (Any): Key within parent dictionary
        key (Any): Key within dictionary
        value (Any): Value to add to set in dictionary

    Returns:
        None

    """
    dict_objs = dictionary.get(parent_key, dict())
    dict_objs[key] = value
    dictionary[parent_key] = dict_objs


def list_distribute_contents_simple(
    input_list: ListTuple, function: Callable[[Any], Any] = lambda x: x
) -> List:
    """Distribute the contents of a list eg. [1, 1, 1, 2, 2, 3] -> [1, 2, 3, 1, 2, 1]. List can contain complex types
    like dictionaries in which case the function can return the appropriate value eg.  lambda x: x[KEY]

    Args:
        input_list (ListTuple): List to distribute values
        function (Callable[[Any], Any]): Return value to use for distributing. Defaults to lambda x: x.

    Returns:
        List: Distributed list

    """
    dictionary = dict()
    for obj in input_list:
        dict_of_lists_add(dictionary, function(obj), obj)
    output_list = list()
    i = 0
    done = False
    while not done:
        found = False
        for key in sorted(dictionary):
            if i < len(dictionary[key]):
                output_list.append(dictionary[key][i])
                found = True
        if found:
            i += 1
        else:
            done = True
    return output_list


def list_distribute_contents(
    input_list: ListTuple, function: Callable[[Any], Any] = lambda x: x
) -> List:
    """Distribute the contents of a list eg. [1, 1, 1, 2, 2, 3] -> [1, 2, 1, 2, 1, 3]. List can contain complex types
    like dictionaries in which case the function can return the appropriate value eg.  lambda x: x[KEY]

    Args:
        input_list (ListTuple): List to distribute values
        function (Callable[[Any], Any]): Return value to use for distributing. Defaults to lambda x: x.

    Returns:
        List: Distributed list

    """

    def riffle_shuffle(piles_list):
        def grouper(n, iterable, fillvalue=None):
            args = [iter(iterable)] * n
            return itertools.zip_longest(fillvalue=fillvalue, *args)

        if not piles_list:
            return []
        piles_list.sort(key=len, reverse=True)
        width = len(piles_list[0])
        pile_iters_list = [iter(pile) for pile in piles_list]
        pile_sizes_list = [
            [pile_position] * len(pile)
            for pile_position, pile in enumerate(piles_list)
        ]
        grouped_rows = grouper(
            width, itertools.chain.from_iterable(pile_sizes_list)
        )
        grouped_columns = itertools.zip_longest(*grouped_rows)
        shuffled_pile = [
            next(pile_iters_list[position])
            for position in itertools.chain.from_iterable(grouped_columns)
            if position is not None
        ]
        return shuffled_pile

    dictionary = dict()
    for obj in input_list:
        dict_of_lists_add(dictionary, function(obj), obj)
    intermediate_list = list()
    for key in sorted(dictionary):
        intermediate_list.append(dictionary[key])
    return riffle_shuffle(intermediate_list)


def extract_list_from_list_of_dict(
    list_of_dict: ListTuple[Dict], key: Any
) -> List:
    """Extract a list by looking up key in each member of a list of dictionaries

    Args:
        list_of_dict (ListTuple[Dict]): List of dictionaries
        key (Any): Key to find in each dictionary

    Returns:
        List: List containing values returned from each dictionary

    """
    result = list()
    for dictionary in list_of_dict:
        result.append(dictionary[key])
    return result


def key_value_convert(
    dictin: Dict,
    keyfn: Callable[[Any], Any] = lambda x: x,
    valuefn: Callable[[Any], Any] = lambda x: x,
    dropfailedkeys: bool = False,
    dropfailedvalues: bool = False,
    exception: Exception = ValueError,
) -> Dict:
    """Convert keys and/or values of dictionary using functions passed in as parameters

    Args:
        dictin (Dict): Input dictionary
        keyfn (Callable[[Any], Any]): Function to convert keys. Defaults to lambda x: x
        valuefn (Callable[[Any], Any]): Function to convert values. Defaults to lambda x: x
        dropfailedkeys (bool): Whether to drop dictionary entries where key conversion fails. Defaults to False.
        dropfailedvalues (bool): Whether to drop dictionary entries where value conversion fails. Defaults to False.
        exception (Exception): The exception to expect if keyfn or valuefn fail. Defaults to ValueError.

    Returns:
        Dict: New dictionary with converted keys and/or values

    """
    dictout = dict()
    for key in dictin:
        try:
            new_key = keyfn(key)
        except exception:
            if dropfailedkeys:
                continue
            new_key = key
        value = dictin[key]
        try:
            new_value = valuefn(value)
        except exception:
            if dropfailedvalues:
                continue
            new_value = value
        dictout[new_key] = new_value
    return dictout


def integer_key_convert(dictin: Dict, dropfailedkeys: bool = False) -> Dict:
    """Convert keys of dictionary to integers

    Args:
        dictin (Dict): Input dictionary
        dropfailedkeys (bool): Whether to drop dictionary entries where key conversion fails. Defaults to False.

    Returns:
        Dict: Dictionary with keys converted to integers

    """
    return key_value_convert(dictin, keyfn=int, dropfailedkeys=dropfailedkeys)


def integer_value_convert(
    dictin: Dict, dropfailedvalues: bool = False
) -> Dict:
    """Convert values of dictionary to integers

    Args:
        dictin (Dict): Input dictionary
        dropfailedvalues (bool): Whether to drop dictionary entries where key conversion fails. Defaults to False.

    Returns:
        Dict: Dictionary with values converted to integers

    """
    return key_value_convert(
        dictin, valuefn=int, dropfailedvalues=dropfailedvalues
    )


def float_value_convert(dictin: Dict, dropfailedvalues: bool = False) -> Dict:
    """Convert values of dictionary to floats

    Args:
        dictin (Dict): Input dictionary
        dropfailedvalues (bool): Whether to drop dictionary entries where key conversion fails. Defaults to False.

    Returns:
        Dict: Dictionary with values converted to floats

    """
    return key_value_convert(
        dictin, valuefn=float, dropfailedvalues=dropfailedvalues
    )


def avg_dicts(dictin1: Dict, dictin2: Dict, dropmissing: bool = True) -> Dict:
    """Create a new dictionary from two dictionaries by averaging values

    Args:
        dictin1 (Dict): First input dictionary
        dictin2 (Dict): Second input dictionary
        dropmissing (bool): Whether to drop keys missing in one dictionary. Defaults to True.

    Returns:
        Dict: Dictionary with values being average of 2 input dictionaries

    """
    dictout = dict()
    for key in dictin1:
        if key in dictin2:
            dictout[key] = (dictin1[key] + dictin2[key]) / 2
        elif not dropmissing:
            dictout[key] = dictin1[key]
    if not dropmissing:
        for key in dictin2:
            if key not in dictin1:
                dictout[key] = dictin2[key]
    return dictout


def read_list_from_csv(
    url: str,
    headers: Union[int, ListTuple[int], ListTuple[str], None] = None,
    dict_form: bool = False,
    **kwargs: Any,
) -> List[ListDict]:
    """Read a list of rows in dict or list form from a csv. The headers argument is
    either a row number or list of row numbers (in case of multi-line headers) to be
    considered as headers (rows start counting at 1), or the actual headers defined as
    a list of strings. If not set, all rows will be treated as containing values.

    Args:
        url (str): URL or path to read from
        headers (Union[int, ListTuple[int], ListTuple[str], None]): Row number of headers. Defaults to None.
        dict_form (bool): Return dict (requires headers parameter) or list for each row. Defaults to False (list)
        **kwargs: Other arguments to pass to Tabulator Stream

    Returns:
        List[ListDict]: List of rows in dict or list form

    """
    if dict_form and headers is None:
        raise ValueError("If dict_form is True, headers must not be None!")
    resource = get_frictionless_resource(url, headers=headers, **kwargs)
    result = list()
    if not dict_form:
        result.append(resource.header)
    for inrow in resource.row_stream:
        if dict_form:
            row = inrow.to_dict()
        else:
            row = inrow.to_list()
        result.append(row)
    resource.close()
    return result


def write_list_to_csv(
    filepath: str,
    rows: List[ListTupleDict],
    headers: Union[int, ListTuple[str], None] = None,
    columns: Union[ListTuple[int], ListTuple[str], None] = None,
    encoding: Optional[str] = None,
) -> None:
    """Write a list of rows in dict or list form to a csv. (The headers argument is
    either a row number (rows start counting at 1), or the actual headers defined as a
    list of strings. If not set, all rows will be treated as containing values.)

    Args:
        filepath (str): Path to write to
        rows (List[ListTupleDict]): List of rows in dict or list form
        headers (Union[int, ListTuple[str], None]): Headers to write. Defaults to None.
        columns (Union[ListTuple[int], ListTuple[str], None]): Columns to write. Defaults to all.
        encoding (Optional[str]): Encoding to use. Defaults to None (infer encoding).

    Returns:
        None

    """
    if len(rows) != 0:
        row = rows[0]
        if isinstance(row, dict):
            has_header = True
            if columns:
                newrows = list()
                for row in rows:
                    newrow = dict()
                    for column in columns:
                        if column in row:
                            newrow[column] = row[column]
                    newrows.append(newrow)
                rows = newrows
                if headers is None:
                    headers = columns
        else:
            if headers is None:
                headers = 1
            if isinstance(headers, int):
                rowno = headers
                headers = rows[rowno - 1]
                rows = rows[rowno:]
            has_header = False
            if columns:
                newrows = list()
                for row in rows:
                    newrow = list()
                    for column in columns:
                        newrow.append(row[column - 1])
                    newrows.append(newrow)
                rows = newrows
        resource = get_frictionless_resource(
            data=rows,
            has_header=has_header,
            headers=headers,
            encoding=encoding,
        )
        resource.write(filepath, format="csv", encoding=encoding)
        resource.close()


def args_to_dict(args: str) -> Dict:
    """Convert command line arguments in a comma separated string to a dictionary

    Args:
        args (str): Command line arguments

    Returns:
        Dict: Dictionary of arguments

    """
    arguments = dict()
    for arg in args.split(","):
        key, value = arg.split("=")
        arguments[key] = value
    return arguments



"""Loader Tests"""
from collections import OrderedDict
from os.path import join

import pytest

from hdx.utilities.loader import (
    LoadError,
    load_and_merge_json,
    load_and_merge_yaml,
    load_json,
    load_json_into_existing_dict,
    load_text,
    load_yaml,
    load_yaml_into_existing_dict,
)
from hdx.utilities.path import temp_dir
from hdx.utilities.saver import save_text


class TestLoader:
    expected_yaml = OrderedDict(
        [
            (
                "hdx_prod_site",
                OrderedDict(
                    [
                        ("url", "https://data.humdata.org/"),
                        ("username", None),
                        ("password", None),
                    ]
                ),
            ),
            (
                "hdx_test_site",
                OrderedDict(
                    [
                        ("url", "https://test-data.humdata.org/"),
                        ("username", "lala"),
                        ("password", "lalala"),
                    ]
                ),
            ),
            (
                "dataset",
                OrderedDict(
                    [("required_fields", ["name", "title", "dataset_date"])]
                ),
            ),
            (
                "resource",
                OrderedDict(
                    [
                        (
                            "required_fields",
                            ["package_id", "name", "description"],
                        )
                    ]
                ),
            ),
            (
                "showcase",
                OrderedDict([("required_fields", ["name", "title"])]),
            ),
            ("param_1", "ABC"),
        ]
    )

    expected_json = OrderedDict(
        [
            (
                "hdx_prod_site",
                OrderedDict(
                    [
                        ("url", "https://data.humdata.org/"),
                        ("username", None),
                        ("password", None),
                    ]
                ),
            ),
            (
                "hdx_test_site",
                OrderedDict(
                    [
                        ("url", "https://test-data.humdata.org/"),
                        ("username", "tumteetum"),
                        ("password", "tumteetumteetum"),
                    ]
                ),
            ),
            (
                "dataset",
                OrderedDict([("required_fields", ["name", "dataset_date"])]),
            ),
            (
                "resource",
                OrderedDict([("required_fields", ["name", "description"])]),
            ),
            ("showcase", OrderedDict([("required_fields", ["name"])])),
            ("my_param", "abc"),
        ]
    )
    text = """  hello
this
is
a
test  

"""  # noqa: W291
    expected_text_strip = """hello
this
is
a
test"""
    expected_text_newlines_to_spaces = """  hello this is a test    """

    def test_load_empty(self, fixturesfolder):
        loaderfolder = join(fixturesfolder, "loader")
        with pytest.raises(LoadError):
            load_text(join(loaderfolder, "empty.yml"))
        with pytest.raises(LoadError):
            load_yaml(join(loaderfolder, "empty.yml"))
        with pytest.raises(LoadError):
            load_json(join(loaderfolder, "empty.json"))

    def test_load_and_merge_yaml(self, configfolder):
        result = load_and_merge_yaml(
            [
                join(configfolder, "hdx_config.yml"),
                join(configfolder, "project_configuration.yml"),
            ]
        )
        assert list(result.items()) == list(TestLoader.expected_yaml.items())

    def test_load_and_merge_json(self, configfolder):
        result = load_and_merge_json(
            [
                join(configfolder, "hdx_config.json"),
                join(configfolder, "project_configuration.json"),
            ]
        )
        assert list(result.items()) == list(TestLoader.expected_json.items())

    def test_load_yaml_into_existing_dict(self, configfolder):
        existing_dict = load_yaml(join(configfolder, "hdx_config.yml"))
        result = load_yaml_into_existing_dict(
            existing_dict, join(configfolder, "project_configuration.yml")
        )
        assert list(result.items()) == list(TestLoader.expected_yaml.items())

    def test_load_json_into_existing_dict(self, configfolder):
        existing_dict = load_json(join(configfolder, "hdx_config.json"))
        result = load_json_into_existing_dict(
            existing_dict, join(configfolder, "project_configuration.json")
        )
        assert list(result.items()) == list(TestLoader.expected_json.items())

    def test_load_file_to_str(self):
        with temp_dir(folder="test_text") as tmpdir:
            text_file = join(tmpdir, "text_file.txt")
            save_text(TestLoader.text, text_file)
            result = load_text(text_file)
            assert result == TestLoader.text
            result = load_text(text_file, strip=True)
            assert result == TestLoader.expected_text_strip
            result = load_text(text_file, replace_newlines=" ")
            assert result == TestLoader.expected_text_newlines_to_spaces
            with pytest.raises(IOError):
                load_text(join(tmpdir, "NOTEXIST.txt"))



"""Downloader Tests"""
import copy
import re
from collections import OrderedDict
from contextlib import contextmanager
from os import remove
from os.path import abspath, join
from shutil import copytree, rmtree
from tempfile import gettempdir

import pytest
from utils import assert_downloaders

from hdx.utilities.base_downloader import DownloadError
from hdx.utilities.downloader import Download
from hdx.utilities.session import SessionError
from hdx.utilities.useragent import UserAgent


@contextmanager
def not_raises(ExpectedException):
    try:
        yield
    except ExpectedException as error:
        raise pytest.fail(f"Raised exception {error} when it should not!")

    except Exception as error:
        raise pytest.fail(f"An unexpected exception {error} raised.")


class TestDownloader:
    downloaderfoldername = "downloader"

    @pytest.fixture(scope="class", autouse=True)
    def useragent(self):
        UserAgent.set_global("test")
        yield
        UserAgent.clear_global()

    @pytest.fixture(scope="class")
    def downloaderfolder(self, fixturesfolder):
        return join(fixturesfolder, self.downloaderfoldername)

    @pytest.fixture(scope="class")
    def fixturefile(self, downloaderfolder):
        return join(downloaderfolder, "extra_params_tree.yml")

    @pytest.fixture(scope="class")
    def fixtureurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/test_data.csv"

    @pytest.fixture(scope="class")
    def fixtureurlexcel(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_data.xlsx"

    @pytest.fixture(scope="class")
    def getfixtureurl(self):
        return "http://httpbin.org/get"

    @pytest.fixture(scope="class")
    def postfixtureurl(self):
        return "http://httpbin.org/post"

    @pytest.fixture(scope="class")
    def fixturenotexistsurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/NOTEXIST.csv"

    @pytest.fixture(scope="class")
    def fixtureprocessurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_csv_processing.csv?a=1"

    @pytest.fixture(scope="class")
    def fixtureprocessurlblank(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/main/tests/fixtures/downloader/test_csv_processing_blanks.csv?a=1"

    def test_get_path_for_url(
        self, tmpdir, fixtureurl, configfolder, downloaderfolder
    ):
        tmpdir = str(tmpdir)
        filename = "test_data.csv"
        path = Download.get_path_for_url(fixtureurl, configfolder)
        assert abspath(path) == abspath(join(configfolder, filename))
        path = Download.get_path_for_url(fixtureurl, downloaderfolder)
        assert abspath(path) == abspath(
            join(downloaderfolder, "test_data3.csv")
        )
        testfolder = join(tmpdir, self.downloaderfoldername)
        rmtree(testfolder, ignore_errors=True)
        copytree(downloaderfolder, testfolder)
        path = Download.get_path_for_url(
            fixtureurl, testfolder, overwrite=True
        )
        assert abspath(path) == abspath(join(testfolder, filename))
        rmtree(testfolder)
        filename = "myfilename.txt"
        path = Download.get_path_for_url(fixtureurl, filename=filename)
        assert abspath(path) == abspath(join(gettempdir(), filename))
        path = Download.get_path_for_url(
            fixtureurl, downloaderfolder, filename
        )
        assert abspath(path) == abspath(join(downloaderfolder, filename))

    def test_init(self, monkeypatch, downloaderfolder):
        basicauthfile = join(downloaderfolder, "basicauth.txt")
        with Download(auth=("u", "p")) as downloader:
            assert downloader.session.auth == ("u", "p")
        basicauth = "Basic dXNlcjpwYXNz"
        with Download(basic_auth=basicauth) as downloader:
            assert downloader.session.auth == ("user", "pass")
        with Download(basic_auth_file=basicauthfile) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
        extraparamsyamltree = join(downloaderfolder, "extra_params_tree.yml")
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
        monkeypatch.setenv("BASIC_AUTH", basicauth)
        with Download() as downloader:
            assert downloader.session.auth == ("user", "pass")
        with pytest.raises(SessionError):
            Download(basic_auth="12345")
        with pytest.raises(SessionError):
            Download(
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        monkeypatch.delenv("BASIC_AUTH")
        with pytest.raises(SessionError):
            Download(
                basic_auth=basicauth,
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(auth=("u", "p"), basic_auth="Basic xxxxxxxxxxxxxxxx")
        extraparamsjson = join(downloaderfolder, "extra_params.json")
        with pytest.raises(SessionError):
            Download(auth=("u", "p"), basic_auth_file=extraparamsjson)
        with pytest.raises(SessionError):
            Download(
                basic_auth="Basic dXNlcjpwYXNz",
                basic_auth_file=extraparamsjson,
            )
        with pytest.raises(SessionError):
            Download(
                auth=("u", "p"),
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(
                basic_auth_file=basicauthfile,
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="missingkey",
            )
        with pytest.raises(IOError):
            Download(basic_auth_file="NOTEXIST")
        extraparamsyaml = join(downloaderfolder, "extra_params.yml")
        test_url = "http://www.lalala.com/lala"
        with Download(
            basic_auth_file=basicauthfile, extra_params_dict={"key1": "val1"}
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            assert downloader.get_full_url(test_url) == f"{test_url}?key1=val1"
        key = "Authorization"
        value = "lala"
        with Download(headers={key: value}) as downloader:
            assert key in downloader.session.headers
            assert downloader.session.headers.get(key) == value
        with Download(extra_params_json=extraparamsjson) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert "param_1=value+1" in full_url
            assert "param_2=value_2" in full_url
            assert "param_3=12" in full_url
        with Download(extra_params_yaml=extraparamsyaml) as downloader:
            full_url = downloader.get_full_url(test_url)
            assert "param1=value1" in full_url
            assert "param2=value+2" in full_url
            assert "param3=10" in full_url
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            full_url = downloader.get_full_url(test_url)
            assert "param1=value+1" in full_url
            assert "param2=value2" in full_url
            assert "param3=11" in full_url
            assert "basic_auth" not in full_url
        monkeypatch.setenv("EXTRA_PARAMS", "param1=value2+3,param2=value3+7")
        with Download(
            extra_params_yaml=extraparamsyamltree, extra_params_lookup="mykey"
        ) as downloader:
            assert downloader.session.auth is None
            full_url = downloader.get_full_url(test_url)
            assert "param1=value2%2B3" in full_url
            assert "param2=value3%2B7" in full_url
            assert "param3=11" not in full_url
            assert "basic_auth" not in full_url
        with Download(
            extra_params_yaml=extraparamsyamltree,
            extra_params_lookup="mykey",
            use_env=False,
        ) as downloader:
            assert downloader.session.auth == ("testuser", "testpass")
            full_url = downloader.get_full_url(test_url)
            assert "param1=value+1" in full_url
            assert "param2=value2" in full_url
            assert "param3=11" in full_url
            assert "basic_auth" not in full_url
        monkeypatch.delenv("EXTRA_PARAMS")
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_json=extraparamsjson,
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_yaml=extraparamsyaml,
            )
        with pytest.raises(SessionError):
            Download(
                extra_params_dict={"key1": "val1"},
                extra_params_yaml=extraparamsyamltree,
                extra_params_lookup="mykey",
            )
        with pytest.raises(IOError):
            Download(extra_params_json="NOTEXIST")
        with pytest.raises(IOError):
            Download(extra_params_yaml="NOTEXIST")
        with not_raises(IOError):
            Download(extra_params_json="NOTEXIST", fail_on_missing_file=False)
            Download(extra_params_yaml="NOTEXIST", fail_on_missing_file=False)
            Download(basic_auth_file="NOTEXIST", fail_on_missing_file=False)

    def test_get_url_for_get(self):
        assert (
            Download.get_url_for_get(
                "http://www.lala.com/hdfa?a=3&b=4",
                OrderedDict([("c", "e"), ("d", "f")]),
            )
            == "http://www.lala.com/hdfa?a=3&b=4&c=e&d=f"
        )
        assert (
            Download.get_url_for_get(
                "http://www.lala.com/hdfa?a=3&b=4", {"c": "e", "d": "f"}
            )
            == "http://www.lala.com/hdfa?a=3&b=4&c=e&d=f"
        )

    def test_get_url_params_for_post(self):
        result = Download.get_url_params_for_post(
            "http://www.lala.com/hdfa?a=3&b=4",
            OrderedDict([("c", "e"), ("d", "f")]),
        )
        assert result[0] == "http://www.lala.com/hdfa"
        assert list(result[1].items()) == list(
            OrderedDict(
                [("a", "3"), ("b", "4"), ("c", "e"), ("d", "f")]
            ).items()
        )
        result = Download.get_url_params_for_post(
            "http://www.lala.com/hdfa?a=3&b=4", {"c": "e", "d": "f"}
        )
        assert result[0] == "http://www.lala.com/hdfa"
        assert list(result[1].items()) == list(
            OrderedDict(
                [("a", "3"), ("b", "4"), ("c", "e"), ("d", "f")]
            ).items()
        )

    def test_hxl_row(self):
        headers = ["a", "b", "c"]
        hxltags = {"b": "#b", "c": "#c"}
        assert Download.hxl_row(headers, hxltags) == ["", "#b", "#c"]
        assert Download.hxl_row(headers, hxltags, dict_form=True) == {
            "a": "",
            "b": "#b",
            "c": "#c",
        }
        assert Download.hxl_row(headers, dict()) == ["", "", ""]
        assert Download.hxl_row([], hxltags) == list()

    def test_setup_stream(
        self, fixtureurl, fixturenotexistsurl, getfixtureurl, postfixtureurl
    ):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup("NOTEXIST://NOTEXIST.csv")
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.setup(fixturenotexistsurl)
        with Download() as downloader:
            downloader.setup(fixtureurl)
            headers = downloader.response.headers
            assert bool(re.match(r"7\d\d", headers["Content-Length"])) is True
        with Download() as downloader:
            downloader.setup(postfixtureurl, post=True)
            headers = downloader.response.headers
            assert bool(re.match(r"4\d\d", headers["Content-Length"])) is True
            downloader.setup(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            downloader.setup(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
            )
            assert list(downloader.get_json()["form"].items()) == list(
                OrderedDict(
                    [("a", "3"), ("c", "2"), ("id", "3"), ("lala", "b")]
                ).items()
            )

    def test_hash_stream(self, fixtureurl):
        with Download() as downloader:
            downloader.setup(fixtureurl)
            md5hash = downloader.hash_stream(fixtureurl)
            assert md5hash == "da9db35a396cca10c618f6795bdb9ff2"

    def test_download_file(
        self,
        tmpdir,
        fixturefile,
        fixtureurl,
        fixturenotexistsurl,
        getfixtureurl,
        postfixtureurl,
    ):
        tmpdir = str(tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file("NOTEXIST://NOTEXIST.csv", folder=tmpdir)
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(fixturenotexistsurl)
        filename = "myfilename.txt"
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(
                fixturefile, folder=tmpdir, path=join(tmpdir, filename)
            )
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download_file(
                fixturefile, filename=filename, path=join(tmpdir, filename)
            )
        with Download() as downloader:
            f = downloader.download_file(fixturefile, folder=tmpdir)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, "extra_params_tree.yml"))
            f = downloader.download_file(fixtureurl, folder=tmpdir)
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, "test_data.csv"))
            f = downloader.download_file(
                fixtureurl, folder=tmpdir, filename=filename
            )
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                fixtureurl, path=join(tmpdir, filename), overwrite=True
            )
            fpath = abspath(f)
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
                folder=tmpdir,
                filename=filename,
            )
            fpath = abspath(f)
            with open(fpath) as fi:
                text = fi.read()
                assert '"id": "10"' in text
                assert '"lala": "a"' in text
                assert '"b": "4"' in text
                assert '"d": "3"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))
            f = downloader.download_file(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
                folder=tmpdir,
                filename=filename,
            )
            fpath = abspath(f)
            with open(fpath) as fi:
                text = fi.read()
                assert '"id": "3"' in text
                assert '"lala": "b"' in text
                assert '"a": "3"' in text
                assert '"c": "2"' in text
            remove(f)
            assert fpath == abspath(join(tmpdir, filename))

    def test_download(
        self,
        fixturefile,
        fixtureurl,
        fixturenotexistsurl,
        getfixtureurl,
        postfixtureurl,
    ):
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download("NOTEXIST://NOTEXIST.csv")
        with pytest.raises(DownloadError), Download() as downloader:
            downloader.download(fixturenotexistsurl)
        with Download() as downloader:
            text = downloader.download_text(fixtureurl)
            assert downloader.get_status() == 200
            assert len(downloader.get_headers()) in (24, 25)
            assert (
                bool(
                    re.match(r"7\d\d", downloader.get_header("Content-Length"))
                )
                is True
            )
            assert (
                text
                == """GWNO,EVENT_ID_CNTY,EVENT_ID_NO_CNTY,EVENT_DATE,YEAR,TIME_PRECISION,EVENT_TYPE,ACTOR1,ALLY_ACTOR_1,INTER1,ACTOR2,ALLY_ACTOR_2,INTER2,INTERACTION,COUNTRY,ADMIN1,ADMIN2,ADMIN3,LOCATION,LATITUDE,LONGITUDE,GEO_PRECISION,SOURCE,NOTES,FATALITIES
615,1416RTA,,18/04/2001,2001,1,Violence against civilians,Police Forces of Algeria (1999-),,1,Civilians (Algeria),Berber Ethnic Group (Algeria),7,17,Algeria,Tizi Ouzou,Beni-Douala,,Beni Douala,36.61954,4.08282,1,Associated Press Online,A Berber student was shot while in police custody at a police station in Beni Douala. He later died on Apr.21.,1
615,2229RTA,,19/04/2001,2001,1,Riots/Protests,Rioters (Algeria),Berber Ethnic Group (Algeria),5,Police Forces of Algeria (1999-),,1,15,Algeria,Tizi Ouzou,Tizi Ouzou,,Tizi Ouzou,36.71183,4.04591,3,Kabylie report,"Riots were reported in numerous villages in Kabylie, resulting in dozens wounded in clashes between protesters and police and significant material damage.",0
615,2230RTA,,20/04/2001,2001,1,Riots/Protests,Protesters (Algeria),Students (Algeria),6,,,0,60,Algeria,Bejaia,Amizour,,Amizour,36.64022,4.90131,1,Crisis Group,Students protested in the Amizour area. At least 3 were later arrested for allegedly insulting gendarmes.,0
615,2231RTA,,21/04/2001,2001,1,Riots/Protests,Rioters (Algeria),Berber Ethnic Group (Algeria),5,Police Forces of Algeria (1999-),,1,15,Algeria,Bejaia,Amizour,,Amizour,36.64022,4.90131,1,Kabylie report,"Rioters threw molotov cocktails, rocks and burning tires at gendarmerie stations in Beni Douala, El-Kseur and Amizour.",0

"""
            )
            downloader.download(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            json = downloader.download_json(
                f"{postfixtureurl}?id=3&lala=b",
                post=True,
                parameters=OrderedDict([("a", "3"), ("c", "2")]),
            )
            assert list(json["form"].items()) == list(
                OrderedDict(
                    [("a", "3"), ("c", "2"), ("id", "3"), ("lala", "b")]
                ).items()
            )
        with Download(rate_limit={"calls": 1, "period": 0.1}) as downloader:
            downloader.download(
                f"{getfixtureurl}?id=10&lala=a",
                post=False,
                parameters=OrderedDict([("b", "4"), ("d", "3")]),
            )
            assert list(downloader.get_json()["args"].items()) == list(
                OrderedDict(
                    [("b", "4"), ("d", "3"), ("id", "10"), ("lala", "a")]
                ).items()
            )
            yaml = downloader.download_yaml(fixturefile)
            assert list(yaml["mykey"].items()) == list(
                OrderedDict(
                    [
                        ("param1", "value 1"),
                        ("param2", "value2"),
                        ("param3", 11),
                        ("basic_auth", "Basic dGVzdHVzZXI6dGVzdHBhc3M="),
                    ]
                ).items()
            )

    def test_download_tabular_key_value(
        self, fixtureurl, fixtureurlexcel, fixtureprocessurl
    ):
        with Download() as downloader:
            result = downloader.download_tabular_key_value(
                fixtureurl,
                file_type="csv",
                delimiter=",",
                skip_initial_space=True,
                infer_types=True,
            )
            assert result == {615: "2231RTA", "GWNO": "EVENT_ID_CNTY"}
            result = downloader.download_tabular_key_value(
                fixtureurlexcel,
                file_type="xlsx",
                sheet="test_data",
                infer_types=True,
            )
            assert result == {615: "2231RTA", "GWNO": "EVENT_ID_CNTY"}
            result = downloader.download_tabular_key_value(
                fixtureprocessurl, headers=3
            )
            assert result == {"coal": "3", "gas": "2"}
            with pytest.raises(DownloadError):
                downloader.download_tabular_key_value(
                    "NOTEXIST://NOTEXIST.csv"
                )

    def test_get_tabular_rows_as_list(self, fixtureprocessurl):
        with Download() as downloader:
            headers, rows = downloader.get_tabular_rows_as_list(
                fixtureprocessurl
            )
            rows = list(rows)
            assert rows == [
                ["la1", "ha1", "ba1", "ma1"],
                ["header1", "header2", "header3", "header4"],
                ["coal", "3", "7.4", "'needed'"],
                ["gas", "2", "6.5", "'n/a'"],
            ]

    def test_get_tabular_rows_as_dict(self, fixtureprocessurl):
        with Download() as downloader:
            headers, rows = downloader.get_tabular_rows_as_dict(
                fixtureprocessurl
            )
            rows = list(rows)
            assert rows == [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                },
                {"la1": "coal", "ha1": "3", "ba1": "7.4", "ma1": "'needed'"},
                {"la1": "gas", "ha1": "2", "ba1": "6.5", "ma1": "'n/a'"},
            ]

    def test_get_tabular_rows(self, fixtureprocessurl, fixtureprocessurlblank):
        with Download() as downloader:
            expected = [
                ["la1", "ha1", "ba1", "ma1"],
                ["header1", "header2", "header3", "header4"],
                ["coal", "3", "7.4", "'needed'"],
                ["gas", "2", "6.5", "'n/a'"],
            ]
            expected_headers = expected[0]
            headers, iterator = downloader.get_tabular_rows(fixtureprocessurl)
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank
            )
            assert headers == expected_headers
            blank_expected = copy.deepcopy(expected[1:])
            blank_expected[2][0] = None
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank, ignore_blank_rows=False
            )
            assert headers == expected_headers
            blank_expected.append([None, None, None, None])
            blank_expected.insert(1, [None, None, None, None])
            assert list(iterator) == blank_expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=1
            )
            assert headers == expected_headers
            assert list(iterator) == expected[1:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=2
            )
            assert headers == expected[1]
            assert list(iterator) == expected[2:]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=[1, 2]
            )
            assert headers == [
                "la1 header1",
                "ha1 header2",
                "ba1 header3",
                "ma1 header4",
            ]
            assert list(iterator) == expected[2:]
            myheaders = ["a", "b", "c", "d"]
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=myheaders
            )
            assert headers == myheaders
            assert list(iterator) == expected
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=1, dict_form=True
            )
            assert headers == expected_headers
            expected_dicts = [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                },
                {"la1": "coal", "ha1": "3", "ba1": "7.4", "ma1": "'needed'"},
                {"la1": "gas", "ha1": "2", "ba1": "6.5", "ma1": "'n/a'"},
            ]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl, headers=3, dict_form=True
            )
            assert headers == expected[2]
            expected_dicts = [
                {"coal": "gas", "3": "2", "7.4": "6.5", "'needed'": "'n/a'"}
            ]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row.insert(2, "lala")
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl,
                headers=3,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            expected_headers = ["coal", "3", "la", "7.4", "'needed'"]
            assert headers == expected_headers
            assert list(iterator) == [["gas", "2", "lala", "6.5", "'n/a'"]]

            def testfn(headers, row):
                row["la"] = "lala"
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurl,
                headers=3,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            expected_dicts[0]["la"] = "lala"
            assert list(iterator) == expected_dicts

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                ignore_blank_rows=False,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            expected_headers = ["la1", "ha1", "la", "ba1", "ma1"]
            assert headers == expected_headers

            expected_dicts = [
                {
                    "la1": "header1",
                    "ha1": "header2",
                    "ba1": "header3",
                    "ma1": "header4",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": None,
                    "ba1": None,
                    "ma1": None,
                    "la": "lala",
                },
                {
                    "la1": "coal",
                    "ha1": "3",
                    "ba1": "7.4",
                    "ma1": "'needed'",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": "2",
                    "ba1": "6.5",
                    "ma1": "'n/a'",
                    "la": "lala",
                },
                {
                    "la1": None,
                    "ha1": None,
                    "ba1": None,
                    "ma1": None,
                    "la": "lala",
                },
            ]
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                ignore_blank_rows=False,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            assert list(iterator) == expected_dicts
            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            del expected_dicts[4]
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            def testfn(headers, row):
                row["la"] = "lala"
                if row["ha1"] == "3":
                    return None
                return row

            headers, iterator = downloader.get_tabular_rows(
                fixtureprocessurlblank,
                headers=1,
                dict_form=True,
                header_insertions=[(2, "la")],
                row_function=testfn,
            )
            assert headers == expected_headers
            del expected_dicts[1]
            assert list(iterator) == expected_dicts

            with pytest.raises(DownloadError):
                downloader.get_tabular_rows(fixtureprocessurl, headers=None)

    def test_download_tabular_rows_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_rows_as_dicts(
                fixtureprocessurl, headers=2
            )
            assert result == {
                "coal": {
                    "header2": "3",
                    "header3": "7.4",
                    "header4": "'needed'",
                },
                "gas": {"header2": "2", "header3": "6.5", "header4": "'n/a'"},
            }
            result = downloader.download_tabular_rows_as_dicts(
                fixtureprocessurl, headers=2, keycolumn=2
            )
            assert result == {
                "2": {"header1": "gas", "header3": "6.5", "header4": "'n/a'"},
                "3": {
                    "header1": "coal",
                    "header3": "7.4",
                    "header4": "'needed'",
                },
            }

    def test_download_tabular_cols_as_dicts(self, fixtureprocessurl):
        with Download() as downloader:
            result = downloader.download_tabular_cols_as_dicts(
                fixtureprocessurl, headers=2
            )
            assert result == {
                "header2": {"coal": "3", "gas": "2"},
                "header3": {"coal": "7.4", "gas": "6.5"},
                "header4": {"coal": "'needed'", "gas": "'n/a'"},
            }
            result = downloader.download_tabular_cols_as_dicts(
                fixtureprocessurl, headers=2, keycolumn=2
            )
            assert result == {
                "header1": {"3": "coal", "2": "gas"},
                "header3": {"3": "7.4", "2": "6.5"},
                "header4": {"3": "'needed'", "2": "'n/a'"},
            }

    def test_get_column_positions(self):
        assert Download.get_column_positions(["a", "b", "c"]) == {
            "a": 0,
            "b": 1,
            "c": 2,
        }

    def test_generate_downloaders(self, downloaders):
        downloader1 = Download.get_downloader()
        downloader2 = Download.get_downloader("test")
        assert_downloaders(downloader1, downloader2, downloaders)



import logging
from os import mkdir
from os.path import join
from shutil import rmtree
from typing import Any, Iterator, List, Optional, Tuple, Union

from slugify import slugify

from .base_downloader import BaseDownload, DownloadError
from .downloader import Download
from .loader import load_json, load_text, load_yaml
from .path import get_filename_extension_from_url
from .saver import save_json, save_text, save_yaml
from .typehint import ListDict, ListTuple

logger = logging.getLogger(__name__)


class Retrieve(BaseDownload):
    """Retrieve class which takes in a Download object and can either download, download
    and save or use previously downloaded and saved data. It also allows the use of a
    static fallback when downloading fails.

    Args:
        downloader (Download): Download object
        fallback_dir (str): Directory containing static fallback data
        saved_dir (str): Directory to save or load downloaded data
        temp_dir (str): Temporary directory for when data is not needed after downloading
        save (bool): Whether to save downloaded data. Defaults to False.
        use_saved (bool): Whether to use saved data. Defaults to False.
        prefix (str): Prefix to add to filenames. Defaults to "".
        delete (bool): Whether to delete saved_dir if save is True. Defaults to True.
        log_level (int): Level at which to log messages. Defaults to logging.INFO.
    """

    retrievers = dict()

    def __init__(
        self,
        downloader: Download,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        prefix: str = "",
        delete: bool = True,
        log_level: int = logging.INFO,
    ):
        self.downloader = downloader
        self.fallback_dir = fallback_dir
        self.saved_dir = saved_dir
        self.temp_dir = temp_dir
        self.save = save
        self.use_saved = use_saved
        self.prefix = prefix
        self.check_flags(saved_dir, save, use_saved, delete)
        self.log_level = log_level

    @staticmethod
    def check_flags(
        saved_dir: str, save: bool, use_saved: bool, delete: bool
    ) -> None:
        """Check flags. Also delete saved_dir if save and delete are True

        Args:
            saved_dir (str): Directory to save or load downloaded data
            save (bool): Whether to save downloaded data
            use_saved (bool): Whether to use saved data
            delete (bool): Whether to delete saved_dir if save is True

        Returns:
            None

        """
        if save:
            if use_saved:
                raise ValueError(
                    "Either the save or use_saved flags can be set to True!"
                )
            if delete:
                rmtree(saved_dir, ignore_errors=True)
                mkdir(saved_dir)

    @staticmethod
    def get_url_logstr(url: str) -> str:
        """Url string that will be logged. It is limited to 100 characters if necessary.

        Args:
            url (str): URL to download

        Returns:
            str: Url string to use in logs

        """
        if len(url) > 100:
            return f"{url[:100]}..."
        return url

    def clone(self, downloader: Download) -> "Retrieve":
        """Clone a given retriever but use the given downloader

        Args:
            downloader (Download): Downloader to use

        Returns:
            Retrieve: Cloned retriever

        """
        return Retrieve(
            downloader,
            fallback_dir=self.fallback_dir,
            saved_dir=self.saved_dir,
            temp_dir=self.temp_dir,
            save=self.save,
            use_saved=self.use_saved,
            prefix=self.prefix,
            delete=False,
        )

    def get_filename(
        self,
        url: str,
        filename: Optional[str] = None,
        possible_extensions: Tuple[str, ...] = tuple(),
        **kwargs: Any,
    ) -> Tuple[str, Any]:
        """Get filename from url and given parameters

        Args:
            url (str): Url from which to get filename
            filename (optional[str]): Filename to use. Defaults to None (infer from url).
            possible_extensions (Tuple[str, ...]): Possible extensions to look for in url
            **kwargs: See below
            format (str): Given extension to look for in url
            file_type (str): Given extension to look for in url

        Returns:
            Tuple[str, Any]: Tuple of (filename, kwargs)

        """
        prefix = kwargs.pop("file_prefix", self.prefix)
        if prefix:
            prefix = f"{prefix}_"
        if filename:
            return f"{prefix}{filename}", kwargs
        filename, extension = get_filename_extension_from_url(
            url, second_last=True, use_query=True
        )
        filename = slugify(filename)
        extensions = list()
        format = kwargs.get("format")
        if format:
            extensions.append(format)
        file_type = kwargs.get("file_type")
        if file_type:
            extensions.append(file_type)
        if possible_extensions:
            extensions.extend(possible_extensions)
        if not extensions:
            return f"{prefix}{filename}{extension}", kwargs
        first_ext = f".{extensions[0].lower()}"
        if not extension:
            return f"{prefix}{filename}{first_ext}", kwargs
        for candidate in extensions:
            if candidate == extension[1:]:
                return f"{prefix}{filename}{extension}", kwargs
        filename = slugify(f"{filename}{extension}")
        return f"{prefix}{filename}{first_ext}", kwargs

    def download_file(
        self,
        url: str,
        filename: Optional[str] = None,
        logstr: Optional[str] = None,
        fallback: bool = False,
        log_level: int = None,
        **kwargs: Any,
    ) -> str:
        """Retrieve file

        Args:
            url (str): URL to download
            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            log_level (int): Level at which to log messages. Overrides level from constructor.
            **kwargs: Parameters to pass to download_file call

        Returns:
            str: Path to downloaded file

        """
        if log_level is None:
            log_level = self.log_level
        filename, kwargs = self.get_filename(url, filename, **kwargs)
        if not logstr:
            logstr = filename
        if self.save:
            folder = self.saved_dir
        else:
            folder = self.temp_dir
        output_path = join(folder, filename)
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.log(log_level, f"Using saved {logstr} in {saved_path}")
            return saved_path
        else:
            try:
                logger.log(
                    log_level,
                    f"Downloading {logstr} from {self.get_url_logstr(url)} into {output_path}",
                )
                return self.downloader.download_file(
                    url, path=output_path, **kwargs
                )
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception(
                    f"{logstr} download failed, using static data {fallback_path}!"
                )
                return fallback_path

    def download_text(
        self,
        url: str,
        filename: Optional[str] = None,
        logstr: Optional[str] = None,
        fallback: bool = False,
        log_level: int = None,
        **kwargs: Any,
    ) -> str:
        """Download text

        Args:
            url (str): URL to download
            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            log_level (int): Level at which to log messages. Overrides level from constructor.
            **kwargs: Parameters to pass to download_text call

        Returns:
            str: The text from the file

        """
        if log_level is None:
            log_level = self.log_level
        filename, kwargs = self.get_filename(url, filename, **kwargs)
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.log(log_level, f"Using saved {logstr} in {saved_path}")
            text = load_text(saved_path)
        else:
            try:
                logger.log(
                    log_level,
                    f"Downloading {logstr} from {self.get_url_logstr(url)}",
                )
                text = self.downloader.download_text(url, **kwargs)
                if self.save:
                    logger.log(log_level, f"Saving {logstr} in {saved_path}")
                    save_text(text, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception(
                    f"{logstr} download failed, using static data {fallback_path}!"
                )
                text = load_text(fallback_path)
        return text

    def download_yaml(
        self,
        url: str,
        filename: Optional[str] = None,
        logstr: Optional[str] = None,
        fallback: bool = False,
        log_level: int = None,
        **kwargs: Any,
    ) -> Any:
        """Retrieve YAML

        Args:
            url (str): URL to download
            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            log_level (int): Level at which to log messages. Overrides level from constructor.
            **kwargs: Parameters to pass to download_yaml call

        Returns:
            Any: The data from the YAML file

        """
        if log_level is None:
            log_level = self.log_level
        filename, kwargs = self.get_filename(
            url, filename, ("yaml", "yml"), **kwargs
        )
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.log(log_level, f"Using saved {logstr} in {saved_path}")
            ryaml = load_yaml(saved_path)
        else:
            try:
                logger.log(
                    log_level,
                    f"Downloading {logstr} from {self.get_url_logstr(url)}",
                )
                ryaml = self.downloader.download_yaml(url, **kwargs)
                if self.save:
                    logger.log(log_level, f"Saving {logstr} in {saved_path}")
                    save_yaml(ryaml, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception(
                    f"{logstr} download failed, using static data {fallback_path}!"
                )
                ryaml = load_yaml(fallback_path)
        return ryaml

    def download_json(
        self,
        url: str,
        filename: Optional[str] = None,
        logstr: Optional[str] = None,
        fallback: bool = False,
        log_level: int = None,
        **kwargs: Any,
    ) -> Any:
        """Retrieve JSON

        Args:
            url (str): URL to download
            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            log_level (int): Level at which to log messages. Overrides level from constructor.
            **kwargs: Parameters to pass to download_json call

        Returns:
            Any: The data from the JSON file

        """
        if log_level is None:
            log_level = self.log_level
        filename, kwargs = self.get_filename(
            url, filename, ("json",), **kwargs
        )
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.log(log_level, f"Using saved {logstr} in {saved_path}")
            rjson = load_json(saved_path)
        else:
            try:
                logger.log(
                    log_level,
                    f"Downloading {logstr} from {self.get_url_logstr(url)}",
                )
                rjson = self.downloader.download_json(url, **kwargs)
                if self.save:
                    logger.log(log_level, f"Saving {logstr} in {saved_path}")
                    save_json(rjson, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception(
                    f"{logstr} download failed, using static data {fallback_path}!"
                )
                rjson = load_json(fallback_path)
        return rjson

    def get_tabular_rows(
        self,
        url: str,
        headers: Union[int, ListTuple[int], ListTuple[str]] = 1,
        dict_form: bool = False,
        filename: Optional[str] = None,
        logstr: Optional[str] = None,
        fallback: bool = False,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[ListDict]]:
        """Returns header of tabular file pointed to by url and an iterator where each
        row is returned as a list or dictionary depending on the dict_rows argument.
        The headers argument is either a row number or list of row numbers (in case of
        multi-line headers) to be considered as headers (rows start counting at 1), or
        the actual headers defined as a list of strings. It defaults to 1.
        The dict_form arguments specifies if each row should be returned as a dictionary
        or a list, defaulting to a list.

        Args:
            url (str): URL or path to read from
            headers (Union[int, ListTuple[int], ListTuple[str]]): Number of row(s) containing headers or list of headers. Defaults to 1.
            dict_form (bool): Return dict or list for each row. Defaults to False (list)
            filename (Optional[str]): Filename of saved file. Defaults to getting from url.
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            **kwargs: Parameters to pass to download_file call

        Returns:
            Tuple[List[str],Iterator[ListDict]]: Tuple (headers, iterator where each row is a list or dictionary)

        """
        path = self.download_file(url, filename, logstr, fallback, **kwargs)
        kwargs.pop("file_prefix", None)
        return self.downloader.get_tabular_rows(
            path, headers, dict_form, **kwargs
        )

    @classmethod
    def generate_retrievers(
        cls,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        ignore: ListTuple[str] = tuple(),
        delete: bool = True,
        **kwargs: Any,
    ) -> None:
        """Generate retrievers. Retrievers are generated from downloaders so
        Download.generate_downloaders() needs to have been called first. Each retriever
        can either download, download and save or use previously downloaded and saved
        data. It also allows the use of a static fallback when downloading fails.

        Args:
            fallback_dir (str): Directory containing static fallback data
            saved_dir (str): Directory to save or load downloaded data
            temp_dir (str): Temporary directory for when data is not needed after downloading
            save (bool): Whether to save downloaded data. Defaults to False.
            use_saved (bool): Whether to use saved data. Defaults to False.
            ignore (ListTuple[str]): Don't generate retrievers for these downloaders
            delete (bool): Whether to delete saved_dir if save is True. Defaults to True.
            **kwargs (Any): Any other arguments to pass.

        Returns:
            None
        """
        cls.check_flags(saved_dir, save, use_saved, delete)
        cls.retrievers = dict()
        for name, downloader in Download.downloaders.items():
            if name in ignore:
                continue
            cls.retrievers[name] = cls(
                downloader,
                fallback_dir,
                saved_dir,
                temp_dir,
                save,
                use_saved,
                delete=False,
                **kwargs,
            )

    @classmethod
    def get_retriever(cls, name: Optional[str] = None) -> "Retrieve":
        """Get a generated retriever given a name. If name is not supplied, the default
        one will be returned.

        Args:
            name (Optional[str]): Name of retriever. Defaults to None (get default).

        Returns:
            Retriever: Retriever object
        """
        return cls.retrievers.get(name, cls.retrievers["default"])



"""Compare Utility Tests"""
from os.path import join

import pytest

from hdx.utilities.compare import assert_files_same, compare_files


class TestCompare:
    @pytest.fixture(scope="class")
    def testfile1(self, fixturesfolder):
        return join(fixturesfolder, "compare", "test_csv_processing.csv")

    @pytest.fixture(scope="class")
    def testfile2(self, fixturesfolder):
        return join(fixturesfolder, "compare", "test_csv_processing2.csv")

    def test_compare_files(self, testfile1, testfile2):
        result = compare_files(testfile1, testfile2)
        assert result == [
            "- coal   ,3      ,7.4    ,'needed'",
            "?         ^\n",
            "+ coal   ,1      ,7.4    ,'notneeded'",
            "?         ^                +++\n",
        ]
        assert_files_same(testfile1, testfile1)



from abc import ABC, abstractmethod
from typing import Any, Iterator, List, Tuple, Union

from .typehint import ListDict, ListTuple


class DownloadError(Exception):
    pass


class BaseDownload(ABC):
    """Base download class with various download operations that subclasses should
    implement.
    """

    def __enter__(self) -> "BaseDownload":
        """
        Allow usage of with

        Returns:
            BaseDownload: Download object

        """
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """
        Subclasses should define this to allow with usage

        Args:
            exc_type (Any): Exception type
            exc_value (Any): Exception value
            traceback (Any): Traceback

        Returns:
            None

        """
        pass

    @abstractmethod
    def download_file(self, url: str, *args: Any, **kwargs: Any) -> str:
        """Download file from url

        Args:
            url (str): URL or path to download
            *args (Any): Positional arguments
            **kwargs (Any): Keyword arguments

        Returns:
            str: Path of downloaded file

        """

    @abstractmethod
    def download_text(self, url: str, *args: Any, **kwargs: Any) -> str:
        """Download text from url

        Args:
            url (str): URL or path to download
            *args (Any): Positional arguments
            **kwargs (Any): Keyword arguments

        Returns:
            str: The text from the file

        """

    @abstractmethod
    def download_yaml(self, url: str, *args: Any, **kwargs: Any) -> Any:
        """Download YAML from url

        Args:
            url (str): URL or path to download
            *args (Any): Positional arguments
            **kwargs (Any): Keyword arguments

        Returns:
            Any: The data from the YAML file

        """

    @abstractmethod
    def download_json(self, url: str, *args: Any, **kwargs: Any) -> Any:
        """Download JSON from url

        Args:
            url (str): URL or path to download
            *args (Any): Positional arguments
            **kwargs (Any): Keyword arguments

        Returns:
            Any: The data from the JSON file

        """

    @abstractmethod
    def get_tabular_rows(
        self,
        url: str,
        headers: Union[int, ListTuple[int], ListTuple[str]] = 1,
        dict_form: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[ListDict]]:
        """Returns header of tabular file pointed to by url and an iterator where each
        row is returned as a list or dictionary depending on the dict_rows argument.
        The headers argument is either a row number or list of row numbers (in case of
        multi-line headers) to be considered as headers (rows start counting at 1), or
        the actual headers defined as a list of strings. It defaults to 1.
        The dict_form arguments specifies if each row should be returned as a dictionary
        or a list, defaulting to a list.

        Args:
            url (str): URL or path to read from
            headers (Union[int, ListTuple[int], ListTuple[str]]): Number of row(s) containing headers or list of headers. Defaults to 1.
            dict_form (bool): Return dict or list for each row. Defaults to False (list)
            *args (Any): Positional arguments
            **kwargs (Any): Keyword arguments

        Returns:
            Tuple[List[str],Iterator[ListDict]]: Tuple (headers, iterator where each row is a list or dictionary)

        """



"""Path Utility Tests"""
import copy
from os.path import exists, join
from shutil import rmtree
from tempfile import gettempdir

import pytest

from hdx.utilities.loader import load_text
from hdx.utilities.path import (
    get_filename_extension_from_url,
    get_filename_from_url,
    get_temp_dir,
    multiple_progress_storing_tempdir,
    progress_storing_tempdir,
    temp_dir,
)


class TestPath:
    @pytest.fixture(scope="class")
    def mytestdir(self):
        return join("haha", "lala")

    @pytest.fixture(scope="class")
    def fixtureurl(self):
        return "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-utilities/master/tests/fixtures/test_data.csv"

    def test_get_temp_dir(self, monkeypatch, mytestdir):
        assert get_temp_dir() == gettempdir()
        assert get_temp_dir("TEST") == join(gettempdir(), "TEST")
        monkeypatch.setenv("TEMP_DIR", mytestdir)
        assert get_temp_dir() == mytestdir
        monkeypatch.delenv("TEMP_DIR")

    def test_temp_dir(self, monkeypatch, mytestdir):
        monkeypatch.setenv("TEMP_DIR", mytestdir)
        with temp_dir() as tempdir:
            assert tempdir == mytestdir
        monkeypatch.delenv("TEMP_DIR")

        tempfolder = "papa"
        expected_dir = join(gettempdir(), tempfolder)

        with temp_dir(tempfolder) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(tempfolder) as tempdir:
                assert tempdir == expected_dir
                raise ValueError("Fail!")
        except ValueError:
            pass
        assert exists(tempdir) is False

        with temp_dir(
            tempfolder, delete_on_success=True, delete_on_failure=True
        ) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(
                tempfolder, delete_on_success=True, delete_on_failure=True
            ) as tempdir:
                assert tempdir == expected_dir
                raise ValueError("Fail!")
        except ValueError:
            pass
        assert exists(tempdir) is False

        with temp_dir(
            tempfolder, delete_on_success=False, delete_on_failure=False
        ) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is True
        rmtree(tempdir)
        try:
            with temp_dir(
                tempfolder, delete_on_success=False, delete_on_failure=False
            ) as tempdir:
                assert tempdir == expected_dir
                raise ValueError("Fail!")
        except ValueError:
            pass
        assert exists(tempdir) is True

        with temp_dir(
            tempfolder, delete_on_success=True, delete_on_failure=False
        ) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is False
        try:
            with temp_dir(
                tempfolder, delete_on_success=True, delete_on_failure=False
            ) as tempdir:
                assert tempdir == expected_dir
                raise ValueError("Fail!")
        except ValueError:
            pass
        assert exists(tempdir) is True
        rmtree(tempdir)

        with temp_dir(
            tempfolder, delete_on_success=False, delete_on_failure=True
        ) as tempdir:
            assert tempdir == expected_dir
        assert exists(tempdir) is True
        rmtree(tempdir)
        try:
            with temp_dir(
                tempfolder, delete_on_success=False, delete_on_failure=True
            ) as tempdir:
                assert tempdir == expected_dir
                raise ValueError("Fail!")
        except ValueError:
            pass
        assert exists(tempdir) is False

    def test_progress_storing_tempdir(self, monkeypatch):
        tempfolder = "papa"
        expected_dir = join(gettempdir(), tempfolder)
        rmtree(expected_dir, ignore_errors=True)
        iterator = [
            {"iso3": "AFG", "name": "Afghanistan"},
            {"iso3": "SDN", "name": "Sudan"},
            {"iso3": "YEM", "name": "Yemen"},
            {"iso3": "ZAM", "name": "Zambia"},
        ]
        expected_batch_file = join(expected_dir, "batch.txt")
        result = list()
        for info, nextdict in progress_storing_tempdir(
            tempfolder, iterator, "iso3"
        ):
            assert info["folder"] == expected_dir
            expected_batch = load_text(expected_batch_file, strip=True)
            result.append(nextdict)
        assert result == iterator
        assert expected_batch == info["batch"]
        assert exists(expected_dir) is False

        monkeypatch.setenv("WHERETOSTART", "iso3=SDN")
        result = list()
        for info, nextdict in progress_storing_tempdir(
            tempfolder, iterator, "iso3"
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == expected_dir
            expected_batch = load_text(expected_batch_file, strip=True)
            result.append(nextdict)
        assert result == iterator[1:]
        assert expected_batch == info["batch"]
        assert exists(expected_dir) is False
        monkeypatch.delenv("WHERETOSTART")

        try:
            for info, nextdict in progress_storing_tempdir(
                tempfolder, iterator, "iso3"
            ):
                if nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        assert exists(expected_dir) is True
        result = list()
        for info, nextdict in progress_storing_tempdir(
            tempfolder, iterator, "iso3"
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == expected_dir
            assert info["batch"] == start_batch
            result.append(nextdict)
        assert result == iterator[2:]
        assert exists(expected_dir) is False

        try:
            for info, nextdict in progress_storing_tempdir(
                tempfolder, iterator, "iso3"
            ):
                if nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        assert exists(expected_dir) is True
        monkeypatch.setenv("WHERETOSTART", "RESET")
        result = list()
        for info, nextdict in progress_storing_tempdir(
            tempfolder, iterator, "iso3"
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == expected_dir
            assert info["batch"] != start_batch
            result.append(nextdict)
        assert result == iterator
        assert exists(expected_dir) is False
        monkeypatch.delenv("WHERETOSTART")

        try:
            for info, nextdict in progress_storing_tempdir(
                tempfolder, iterator, "iso3"
            ):
                if nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        assert exists(expected_dir) is True
        monkeypatch.setenv("WHERETOSTART", "iso3=SDN")
        result = list()
        for info, nextdict in progress_storing_tempdir(
            tempfolder, iterator, "iso3"
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == expected_dir
            assert info["batch"] == start_batch
            result.append(nextdict)
        assert result == iterator[1:]
        assert exists(expected_dir) is False
        monkeypatch.delenv("WHERETOSTART")

        try:
            for info, nextdict in progress_storing_tempdir(
                tempfolder, iterator, "iso3"
            ):
                if nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        monkeypatch.setenv("WHERETOSTART", "iso3=NOTFOUND")
        found = False
        for _ in progress_storing_tempdir(tempfolder, iterator, "iso3"):
            found = True
        assert found is False
        assert exists(expected_dir) is True
        batch = load_text(expected_batch_file, strip=True)
        assert batch == start_batch
        monkeypatch.delenv("WHERETOSTART")

        monkeypatch.setenv("WHERETOSTART", "NOTFOUND=SDN")
        found = False
        for _ in progress_storing_tempdir(tempfolder, iterator, "iso3"):
            found = True
        assert found is False
        assert exists(expected_dir) is True
        batch = load_text(expected_batch_file, strip=True)
        assert batch == start_batch
        monkeypatch.delenv("WHERETOSTART")

        rmtree(expected_dir, ignore_errors=True)

    def test_multiple_progress_storing_tempdir(self, monkeypatch):
        tempfolder = "gaga"
        expected_dir = join(gettempdir(), tempfolder)
        rmtree(expected_dir, ignore_errors=True)
        iterator1 = [{"emergency_id": "911"}]
        iterator2 = [
            {"iso3": "AFG", "name": "Afghanistan"},
            {"iso3": "SDN", "name": "Sudan"},
            {"iso3": "YEM", "name": "Yemen"},
            {"iso3": "ZAM", "name": "Zambia"},
        ]
        iterators = [iterator1, iterator2]
        keys = ["emergency_id", "iso3"]
        results = list()
        for result in multiple_progress_storing_tempdir(
            tempfolder, iterators, keys, "1234"
        ):
            results.append(copy.deepcopy(result))
        expected_results = [
            (
                0,
                {
                    "folder": "/tmp/gaga/0",
                    "batch": "1234",
                    "progress": "emergency_id=911",
                },
                {"emergency_id": "911"},
            ),
            (
                1,
                {
                    "folder": "/tmp/gaga/1",
                    "batch": "1234",
                    "progress": "iso3=AFG",
                },
                {"iso3": "AFG", "name": "Afghanistan"},
            ),
            (
                1,
                {
                    "folder": "/tmp/gaga/1",
                    "batch": "1234",
                    "progress": "iso3=SDN",
                },
                {"iso3": "SDN", "name": "Sudan"},
            ),
            (
                1,
                {
                    "folder": "/tmp/gaga/1",
                    "batch": "1234",
                    "progress": "iso3=YEM",
                },
                {"iso3": "YEM", "name": "Yemen"},
            ),
            (
                1,
                {
                    "folder": "/tmp/gaga/1",
                    "batch": "1234",
                    "progress": "iso3=ZAM",
                },
                {"iso3": "ZAM", "name": "Zambia"},
            ),
        ]
        assert results == expected_results
        assert exists(expected_dir) is False
        results = list()
        try:
            for result in multiple_progress_storing_tempdir(
                tempfolder, iterators, keys
            ):
                results.append(copy.deepcopy(result))
                i, info, nextdict = result
                if "iso3" in nextdict and nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        for result in expected_results:
            result[1]["batch"] = start_batch
        assert results == expected_results[:4]
        assert exists(expected_dir) is True
        result = list()
        for _, info, nextdict in multiple_progress_storing_tempdir(
            tempfolder, iterators, keys
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == join(expected_dir, "1")
            assert info["batch"] == start_batch
            result.append(nextdict)
        assert result == iterator2[2:]
        assert exists(expected_dir) is False

        try:
            for _, info, nextdict in multiple_progress_storing_tempdir(
                tempfolder, iterators, keys
            ):
                if "iso3" in nextdict and nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        for result in expected_results:
            result[1]["batch"] = start_batch
        assert exists(expected_dir) is True
        monkeypatch.setenv("WHERETOSTART", "RESET")
        results = list()
        for result in multiple_progress_storing_tempdir(
            tempfolder, iterators, keys, "1234"
        ):
            results.append(copy.deepcopy(result))
        for result in expected_results:
            result[1]["batch"] = "1234"
        assert results == expected_results
        assert exists(expected_dir) is False
        monkeypatch.delenv("WHERETOSTART")

        try:
            for _, info, nextdict in multiple_progress_storing_tempdir(
                tempfolder, iterators, keys
            ):
                if "iso3" in nextdict and nextdict["iso3"] == "YEM":
                    start_batch = info["batch"]
                    raise ValueError("Problem!")
        except ValueError:
            pass
        for result in expected_results:
            result[1]["batch"] = start_batch
        assert exists(expected_dir) is True
        monkeypatch.setenv("WHERETOSTART", "iso3=SDN")
        result = list()
        for _, info, nextdict in multiple_progress_storing_tempdir(
            tempfolder, iterators, keys
        ):
            assert exists(info["folder"]) is True
            assert info["folder"] == join(expected_dir, "1")
            assert info["batch"] == start_batch
            result.append(nextdict)
        assert result == iterator2[1:]
        assert exists(expected_dir) is False
        monkeypatch.delenv("WHERETOSTART")
        rmtree(expected_dir, ignore_errors=True)

    def test_get_filename_extension_from_url(self, fixtureurl):
        filename = get_filename_from_url(
            "http://test.com/test.csv", second_last=True
        )
        assert filename == "test.csv"
        filename = get_filename_from_url("http://test.com/test/test.csv", True)
        assert filename == "test_test.csv"
        filename = get_filename_from_url(
            "https://globalhealth5050.org/?_covid-data=dataset-fullvars&_extype=csv",
            second_last=True,
        )
        assert filename == "covid-data-dataset-fullvars-extype-csv"
        filename = get_filename_from_url(fixtureurl)
        assert filename == "test_data.csv"
        filename = get_filename_from_url(fixtureurl, second_last=True)
        assert filename == "fixtures_test_data.csv"
        filename, extension = get_filename_extension_from_url(fixtureurl)
        assert filename == "test_data"
        assert extension == ".csv"
        filename, extension = get_filename_extension_from_url(
            fixtureurl, second_last=True
        )
        assert filename == "fixtures_test_data"
        assert extension == ".csv"



"""Loading utilities for YAML, JSON etc."""

import json
from os import linesep
from typing import Any, Dict, Optional

from ruamel.yaml import YAML

from .dictandlist import merge_dictionaries, merge_two_dictionaries
from .typehint import ListTuple


class LoadError(Exception):
    pass


def load_text(
    path: str,
    encoding: str = "utf-8",
    strip: bool = False,
    replace_newlines: Optional[str] = None,
) -> str:
    """
    Load file into a string removing newlines

    Args:
        path (str): Path to file
        encoding (str): Encoding of file. Defaults to utf-8.
        strip (bool): Whether to strip whitespace from start and end. Defaults to False.
        replace_newlines (Optional[str]): String with which tp replace newlines. Defaults to None (don't replace).

    Returns:
        str: String contents of file

    """
    with open(path, encoding=encoding) as f:
        string = f.read()
        if replace_newlines is not None:
            string = string.replace(linesep, replace_newlines)
        if strip:
            string = string.strip()
    if not string:
        raise LoadError(f"{path} file is empty!")
    return string


def load_yaml(path: str, encoding: str = "utf-8") -> Any:
    """Load YAML file into an ordered dictionary

    Args:
        path (str): Path to YAML file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Any: The data from the YAML file
    """
    with open(path, encoding=encoding) as f:
        yaml = YAML()
        yamlobj = yaml.load(f.read())
    if not yamlobj:
        raise (LoadError(f"YAML file: {path} is empty!"))
    return yamlobj


def load_json(path: str, encoding: str = "utf-8") -> Dict:
    """Load JSON file into an ordered dictionary (dict for Python 3.7+)

    Args:
        path (str): Path to JSON file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Any: The data from the JSON file
    """
    with open(path, encoding=encoding) as f:
        jsonobj = json.loads(f.read())
    if not jsonobj:
        raise (LoadError(f"JSON file: {path} is empty!"))
    return jsonobj


def load_and_merge_yaml(
    paths: ListTuple[str], encoding: str = "utf-8"
) -> Dict:
    """Load multiple YAML files that are in dictionary form and merge into one
    dictionary

    Args:
        paths (ListTuple[str]): Paths to YAML files
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: Dictionary of merged YAML files

    """
    configs = [load_yaml(path, encoding=encoding) for path in paths]
    return merge_dictionaries(configs)


def load_and_merge_json(
    paths: ListTuple[str], encoding: str = "utf-8"
) -> Dict:
    """Load multiple JSON files that are in dictionary form and merge into one
    dictionary

    Args:
        paths (ListTuple[str]): Paths to JSON files
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: Dictionary of merged JSON files

    """
    configs = [load_json(path, encoding=encoding) for path in paths]
    return merge_dictionaries(configs)


def load_yaml_into_existing_dict(
    data: dict, path: str, encoding: str = "utf-8"
) -> Dict:
    """Merge YAML file that is in dictionary form into existing dictionary

    Args:
        data (dict): Dictionary to merge into
        path (str): YAML file to load and merge
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        Dict: YAML file merged into dictionary
    """
    yamldict = load_yaml(path, encoding=encoding)
    return merge_two_dictionaries(data, yamldict)


def load_json_into_existing_dict(
    data: dict, path: str, encoding: str = "utf-8"
) -> Dict:
    """Merge JSON file that is in dictionary form into existing dictionary

    Args:
        data (dict): Dictionary to merge into
        path (str): JSON file to load and merge
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        dict: JSON file merged into dictionary
    """
    jsondict = load_json(path, encoding=encoding)
    return merge_two_dictionaries(data, jsondict)



"""User Agent Tests"""
from os.path import join

import pytest

from hdx.utilities import __version__
from hdx.utilities.loader import LoadError
from hdx.utilities.useragent import UserAgent, UserAgentError


class TestUserAgent:
    @pytest.fixture(scope="class")
    def user_agent_config_yaml(self, configfolder):
        return join(configfolder, "user_agent_config.yml")

    @pytest.fixture(scope="class")
    def user_agent_config2_yaml(self, configfolder):
        return join(configfolder, "user_agent_config2.yml")

    @pytest.fixture(scope="class")
    def user_agent_config3_yaml(self, configfolder):
        return join(configfolder, "user_agent_config3.yml")

    @pytest.fixture(scope="class")
    def empty_yaml(self, configfolder):
        return join(configfolder, "empty.yml")

    @pytest.fixture(scope="class")
    def user_agent_config_wrong_yaml(self, configfolder):
        return join(configfolder, "user_agent_config_wrong.yml")

    def test_user_agent(
        self,
        monkeypatch,
        user_agent_config_yaml,
        user_agent_config2_yaml,
        user_agent_config3_yaml,
        empty_yaml,
        user_agent_config_wrong_yaml,
    ):
        assert (
            UserAgent.get(user_agent_config_yaml=user_agent_config_yaml)
            == f"lala:HDXPythonUtilities/{__version__}-myua"
        )
        assert (
            UserAgent.get(user_agent_config_yaml=user_agent_config2_yaml)
            == f"HDXPythonUtilities/{__version__}-myuseragent"
        )
        assert (
            UserAgent.get(
                user_agent_config_yaml=user_agent_config3_yaml,
                user_agent_lookup="lookup",
            )
            == f"HDXPythonUtilities/{__version__}-mylookupagent"
        )
        assert (
            UserAgent.get(user_agent="my_ua", preprefix="papa")
            == f"papa:HDXPythonUtilities/{__version__}-my_ua"
        )
        UserAgent.set_global(
            user_agent_config_yaml=user_agent_config3_yaml,
            user_agent_lookup="lookup2",
        )
        assert (
            UserAgent.get()
            == f"HDXPythonUtilities/{__version__}-mylookupagent2"
        )
        UserAgent.clear_global()
        with pytest.raises(UserAgentError):
            UserAgent.get(
                user_agent_config_yaml=user_agent_config3_yaml,
                user_agent_lookup="fail",
            )
        with pytest.raises(LoadError):
            UserAgent.get(user_agent_config_yaml=empty_yaml)
        with pytest.raises(UserAgentError):
            UserAgent.get(user_agent_config_yaml=user_agent_config_wrong_yaml)
        with pytest.raises(UserAgentError):
            UserAgent.get(user_agent_config_yaml="")
        with pytest.raises(UserAgentError):
            UserAgent.get()
        with pytest.raises(UserAgentError):
            UserAgent._load(prefix="", user_agent_config_yaml="")
        my_user_agent = "lala"
        monkeypatch.setenv("USER_AGENT", my_user_agent)
        assert (
            UserAgent.get()
            == f"HDXPythonUtilities/{__version__}-{my_user_agent}"
        )
        my_preprefix = "haha"
        monkeypatch.setenv("PREPREFIX", my_preprefix)
        assert (
            UserAgent.get()
            == f"{my_preprefix}:HDXPythonUtilities/{__version__}-{my_user_agent}"
        )
        my_prefix = f"HDXPythonLibrary/{__version__}"
        assert (
            UserAgent.get(prefix=my_prefix)
            == f"{my_preprefix}:{my_prefix}-{my_user_agent}"
        )
        UserAgent.clear_global()



"""Email Tests"""
from os.path import join

import pytest

from hdx.utilities.email import Email, EmailConfigurationError


class TestEmail:
    @pytest.fixture(scope="class")
    def email_json(self, configfolder):
        return join(configfolder, "hdx_email_configuration.json")

    @pytest.fixture(scope="class")
    def email_yaml(self, configfolder):
        return join(configfolder, "hdx_email_configuration.yml")

    def test_mail(self, mocksmtp):
        smtp_initargs = {
            "host": "localhost",
            "port": 123,
            "local_hostname": "mycomputer.fqdn.com",
            "timeout": 3,
            "source_address": ("machine", 456),
        }
        username = "user@user.com"
        password = "pass"
        email_config_dict = {
            "connection_type": "ssl",
            "username": username,
            "password": password,
        }
        email_config_dict.update(smtp_initargs)

        recipients = ["larry@gmail.com", "moe@gmail.com", "curly@gmail.com"]
        cc = ["tweedledum@gmail.com", "tweedledee@gmail.com"]
        bcc = ["theinvisibleman@gmail.com", "ghost@gmail.com"]
        subject = "hello"
        text_body = "hello there"
        html_body = """\
<html>
  <head></head>
  <body>
    <p>Hi!<br>
       How are you?<br>
       Here is the <a href="https://www.python.org">link</a> you wanted.
    </p>
  </body>
</html>
        """
        sender = "me@gmail.com"
        mail_options = ["a", "b"]
        rcpt_options = [1, 2]

        with Email(email_config_dict=email_config_dict) as email:
            email.send(
                recipients,
                subject,
                text_body,
                sender=sender,
                cc=cc,
                bcc=bcc,
                mail_options=mail_options,
                rcpt_options=rcpt_options,
            )
            assert email.server.type == "smtpssl"
            assert email.server.initargs == smtp_initargs
            assert email.server.username == username
            assert email.server.password == password
            assert email.server.sender == sender
            assert email.server.recipients == recipients + cc + bcc
            assert (
                email.server.msg
                == """Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Subject: hello
From: me@gmail.com
To: larry@gmail.com, moe@gmail.com, curly@gmail.com
Cc: tweedledum@gmail.com, tweedledee@gmail.com

hello there"""
            )
            assert email.server.send_args == {
                "mail_options": ["a", "b"],
                "rcpt_options": [1, 2],
            }
            email.send(
                recipients,
                subject,
                text_body,
                html_body=html_body,
                sender=sender,
                mail_options=mail_options,
                rcpt_options=rcpt_options,
            )
            assert "Content-Type: multipart/alternative;" in email.server.msg
            assert (
                """\
MIME-Version: 1.0
Subject: hello
From: me@gmail.com
To: larry@gmail.com, moe@gmail.com, curly@gmail.com"""
                in email.server.msg
            )
            assert (
                """\
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

hello there"""
                in email.server.msg
            )
            assert (
                """\
Content-Type: text/html; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

<html>
  <head></head>
  <body>
    <p>Hi!<br>
       How are you?<br>
       Here is the <a href="https://www.python.org">link</a> you wanted.
    </p>
  </body>
</html>"""
                in email.server.msg
            )
            email.send(
                recipients,
                subject,
                text_body,
                mail_options=mail_options,
                rcpt_options=rcpt_options,
            )
            assert email.server.sender == username
            email.send(
                "larry@gmail.com",
                subject,
                text_body,
                html_body=html_body,
                sender=sender,
                cc="moe@gmail.com",
                bcc="curly@gmail.com",
                mail_options=mail_options,
                rcpt_options=rcpt_options,
            )
            assert email.server.recipients == recipients

    def test_json(self, mocksmtp, email_json):
        with Email(email_config_json=email_json) as email:
            email.connect()
            assert email.server.type == "lmtp"

    def test_yaml(self, mocksmtp, email_yaml):
        with Email(email_config_yaml=email_yaml) as email:
            email.connect()
            assert email.server.type == "smtp"

    def test_fail(self, mocksmtp, email_json, email_yaml):
        email_config_dict = {
            "connection_type": "ssl",
        }
        default_email_config_yaml = Email.default_email_config_yaml
        Email.default_email_config_yaml = "NOT EXIST"
        with pytest.raises(IOError):
            with Email() as _:
                pass
        Email.default_email_config_yaml = default_email_config_yaml
        with pytest.raises(EmailConfigurationError):
            with Email(
                email_config_dict=email_config_dict,
                email_config_json=email_json,
            ) as _:
                pass
        with pytest.raises(EmailConfigurationError):
            with Email(
                email_config_dict=email_config_dict,
                email_config_yaml=email_yaml,
            ) as _:
                pass



"""Encoding utilities"""

import base64


def str_to_base64(string: str) -> str:
    """
    Base 64 encode string

    Args:
        string (str): String to encode

    Returns:
        str: Base 64 encoded string

    """
    return base64.urlsafe_b64encode(string.encode("utf-8")).decode("utf-8")


def base64_to_str(bstring: str) -> str:
    """
    Base 64 decode string

    Args:
        bstring (str): Base 64 encoded string to encode

    Returns:
        str: Decoded string

    """
    return base64.urlsafe_b64decode(bstring.encode("utf-8")).decode("utf-8")



"""HTML parsing utilities"""

import logging
from typing import Any, Dict, List, Optional

try:
    from bs4 import BeautifulSoup, Tag
except ImportError:
    BeautifulSoup = None
    Tag = None

from .downloader import Download

logger = logging.getLogger(__name__)


if BeautifulSoup is not None:

    def get_soup(
        url: str,
        downloader: Download = None,
        user_agent: Optional[str] = None,
        user_agent_config_yaml: Optional[str] = None,
        user_agent_lookup: Optional[str] = None,
        **kwargs: Any,
    ) -> BeautifulSoup:
        """
        Get BeautifulSoup object for a url. Requires either global user agent to be set or appropriate user agent
        parameter(s) to be completed.

        Args:
            url (str): url to read
            downloader (Download): Download object. Defaults to creating a Download object with given user agent values.
            user_agent (Optional[str]): User agent string. HDXPythonUtilities/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.

        Returns:
            BeautifulSoup: The BeautifulSoup object for a url

        """
        if not downloader:
            downloader = Download(
                user_agent, user_agent_config_yaml, user_agent_lookup, **kwargs
            )
        response = downloader.download(url)
        return BeautifulSoup(response.text, "html.parser")

    def get_text(tag: Tag) -> str:
        """
        Get text of tag stripped of leading and trailing whitespace and newlines and with &nbsp replaced with space

        Args:
            tag (Tag): BeautifulSoup tag

        Returns:
            str: Text of tag stripped of leading and trailing whitespace and newlines and with &nbsp replaced with space

        """
        return tag.get_text().strip(" \t\n\r").replace("\xa0", " ")

    def extract_table(tabletag: Tag) -> List[Dict]:
        """
        Extract HTML table as list of dictionaries

        Args:
            tabletag (Tag): BeautifulSoup tag

        Returns:
            str: Text of tag stripped of leading and trailing whitespace and newlines and with &nbsp replaced with space

        """
        theadtag = tabletag.find_next("thead")

        headertags = theadtag.find_all("th")
        if len(headertags) == 0:
            headertags = theadtag.find_all("td")
        headers = []
        for tag in headertags:
            headers.append(get_text(tag))

        tbodytag = tabletag.find_next("tbody")
        trtags = tbodytag.find_all("tr")

        table = list()
        for trtag in trtags:
            row = dict()
            tdtags = trtag.find_all("td")
            for i, tag in enumerate(tdtags):
                row[headers[i]] = get_text(tag)
            table.append(row)
        return table
